/*
 * MacGate - LocalTalk to IP GateWay for Linux.
 *      Written By: Jay Schulist <Jay.Schulist@spacs.k12.wi.us>
 *                  Copyright (C) 1997 Jay Schulist
 *
 * Idea Based On:
 *      ipddpd - The IP to LocalTalk Daemon by Bradford W. Johnson.
 *
 * This software may be used and distributed according to the terms
 * of the GNU Public License, incorporated herein by reference.
 */

/*
 *	19970808	Jay Schulist	Added Expire Devices/Node Map 
 *					entries. (Not working yet)
 *					Removed need to specify IP Gateway.
 *					Optimized MacGate a ton.
 */

/*
 * TODO:
 * 1. This is somewhat of a mess and needs cleaning up.
 * 2. This may not work with multiple gateways.. start test'in it.
 * 3. Remove IPFWADM rule for devices when MacGate closes.
 *
 *	!! Need to figure out how to handle routes. 
 *	   Maybe reimplement ExpireNodes(); :(
 */

#include "MacGate.h"    /* Our Stuff. */

/* Global variables. */
int    	SockIP, SockLTIP;	/* Socket file descriptors. */
int	NodesUp;		/* Number of Macs currently active. */

/* 
 * Information of Local Machine (Where MacGate runs). 
 */
struct at_addr MacGateNetNode;         	/* Default LT device's Net:Node pair. */
struct in_addr MacGateIP;		/* IP of Base Ethernet Device. */

/* Index to functions, as function prototypes. */
/* Initalization functions. */
void MacGateInit 	(void);
void MacGateInfo 	(void);
void MacGateReg  	(void);
void SigHandler  	(int sig);

/* Sending and recieving operations. */
void MacGateListen	(void);
void SendToLTalk 	(struct ltip_list **LTNodes, char buff[], int bytes);
void SendToInet 	(struct ltip_list *LTNodes, char buff[], int bytes);
void UpdateNodeMap      (struct ltip_list **LTNodes, char buff[]);

/* Node Map functions. */
struct ltip_list *ReturnNodeByIP (struct ltip_list **LTNodes, __u32 IPnumber);
void NewEtherNode  (struct ltip_list **LTNodes,__u32 LTNewNodeIP,char name[]);
void AddNode 	   (struct ltip_list **LTNodes, struct ltip_list *LTNewNode);
int  AddNodeBefore (struct ltip_list LTNodeOne, struct ltip_list LTNodeTwo);
void ExpireNodes        (struct ltip_list **LTNodes);
void DeleteNode         (struct ltip_list **LTNodes, struct ltip_list OldNode);
void EtherNodeDown 	(char name[]);
void AllEtherDown	(void);

/* Functions that live in Misc.c */
int  GetHWaddrByDev 	(char name[], struct sockaddr *hwaddr);
void PrintList          (struct ltip_list *ltip_nodes);
char *in_ntoa		(__u32 in);

int main (void)
{
	MacGateInit();		/* Initalization stuff. */

	/* Safeway to shutdown MacGate. */
	signal(SIGINT,  SigHandler);
	signal(SIGKILL, SigHandler);
	signal(SIGHUP,  SigHandler);
	signal(SIGTERM, SigHandler);

	switch (fork()) 
	{
        	case 0:
                        MacGateListen();        /* Listen to our sockets. */
                        break;
                case -1:
                        perror("MacGate: fork()");
                        exit(1);
                default:
                        exit(0);
                        break;
        }
        return 0;
}

/*
 * Start up the sockets and get out IP/HW numbers from the
 * devices. Also register our NBP names.
 */
void MacGateInit (void)
{
	printf("MacGate, IP <-> LocalTalk Gateway. v0.02 - 08/08/1997\n");
	printf("Written by Jay Schulist <Jay.Schulist@spacs.k12.wi.us>\n");

	if(getuid()) 
	{
                printf("MacGate: Must be superuser.");
                exit(1);
        }

	MacGateInfo();	/* Get our GateWay's information. */
        MacGateReg();   /* Register out NBP names. */

        /* Inet sockets we listen to (SOCK_PACKET). */
        SockIP	   = socket(AF_INET, SOCK_PACKET, htons(ETH_P_IP));
	SockLTIP   = socket(AF_INET, SOCK_PACKET, htons(ETH_P_LOCALTALK));
        if((SockIP | SockLTIP) < 0)
        {
                perror("MacGate: MacGateInit - socket()");
                exit(1);
        }

        return;
}

/*
 * Get our Net:Node number from Linux LocalTalk device,
 * also get the Ethernet device information along with
 * the IP Gateway information.
 */
void MacGateInfo (void)
{
	struct sockaddr_in *s_in;
	struct sockaddr hwaddr;
	struct ifreq ifr;
        int SockFD;

        SockFD = socket(AF_INET, SOCK_DGRAM, 0);
	if(SockFD < 0)
	{
		perror("MacGate: MacGateInfo - socket(SockFD)");
		exit(1);
	}

	/*
	 * Get Base Device info and copy it over. Odd eh? 
	 */
        strcpy(ifr.ifr_name, ETH_DEV);
        if (ioctl(SockFD, SIOCGIFADDR, &ifr) < 0) perror("SIOCGIFADDR");
        if (ioctl(SockFD, SIOCGIFADDR, &ifr) < 0) perror("SIOCGIFADDR");
        s_in = (struct sockaddr_in *)&ifr.ifr_addr;
        MacGateIP.s_addr = s_in->sin_addr.s_addr;

	/*
	 * Get the node address from the LocalTalk Card. 
	 */
	memset(&hwaddr, '\0', sizeof(hwaddr));
	GetHWaddrByDev(LT_DEV, &hwaddr);
	MacGateNetNode.s_net  = 0;	/* Just to be safe. */
	MacGateNetNode.s_node = hwaddr.sa_data[0];
	strcpy(ifr.ifr_name, LT_DEV);	/* IP is usless, Just there. */
	if (ioctl(SockFD, SIOCGIFADDR, &ifr) < 0) perror("SIOCGIFADDR");

	/* Tell the user what devices we are using. */
	printf("Using Ethernet %s, IP %s and Localtalk %s, Node %d\n", 
                ETH_DEV, inet_ntoa(MacGateIP), LT_DEV, MacGateNetNode.s_node);

        return;
}

/*
 * Register our Ethernet IP Gateway as the IPGATEWAY and IPADDRESS.
 */
void MacGateReg (void)
{
	ATP atp;	/* ATP handle. */
	int SockLTalk;	/* Socket descriptor. */

	/* Socket into kernel Appletalk layer. */
	SockLTalk = socket(AF_APPLETALK, SOCK_DGRAM, 0);
	if(SockLTalk < 0)
	{
                perror("MacGate: MacGateInit - socket()");
                exit(1);
        }	

	atp = atp_open(MAC_IP_SOCKET); 	/* IP DDP Port (72). */
        if(atp == NULL)
        {
                perror("MacGate: MacGateReg - atp_open()");
                exit(1);
        }

        if (nbp_rgstr(atp_sockaddr(atp),inet_ntoa(MacGateIP),"IPADDRESS","*")<0)
        {
                printf("MacGate: MacGateReg - IPADDRESS nbp_rgstr failed.\n");
                exit(1);
        }
	else
	{
		syslog(LOG_INFO,"MacGate: Registered %s as our IPADDRESS.\n",
			inet_ntoa(MacGateIP));
	}

        if (nbp_rgstr(atp_sockaddr(atp),inet_ntoa(MacGateIP),"IPGATEWAY","*")<0)
        {
                printf("MacGate: MacGateReg - IPGATEWAY nbp_rgstr failed.\n");
                exit(1);
        }
	else
	{
		syslog(LOG_INFO,"MacGate: Registered %s as our IPGATEWAY.\n",
			inet_ntoa(MacGateIP));
	}

        return;
}

/*
 * Unregister our IPGATEWAY/IPADDRESS and bring down all our interfaces.
 * FIXME: Needs some serious reworking.
 */
void SigHandler (int sig)
{
        nbp_unrgstr(inet_ntoa(MacGateIP), "IPGATEWAY", "*");
        nbp_unrgstr(inet_ntoa(MacGateIP), "IPADDRESS", "*");
	AllEtherDown();
	syslog(LOG_INFO, "MacGate: Shutdown cleanly");

        exit(0);
}

/*
 * MacGateListen - Here we do all of the work. 
 */
void MacGateListen (void)
{
	fd_set HearData;	
        struct sockaddr from;
        char buf[1024];                /* Large just to be safe. */
        int from_len, bytes, no, max;

        struct ltip_list *LTNodes;     /* Linked list of nodes we gateway for */
        LTNodes = NULL;                /* Set list to empty. */

        for(;;)		/* Here we start our big loop. */
        {
		do /* Listen for data comming in. */
		{
			FD_ZERO(&HearData);
                	FD_SET(SockIP, &HearData);
			FD_SET(SockLTIP, &HearData);
			max = MAX(SockIP, SockLTIP);
		} while((no = select(max + 1, &HearData,
			(fd_set *) 0, (fd_set *) 0,
                	(struct timeval *) 0)) < 0 && errno == EINTR);

		if(no <= 0)
		{
                	syslog(LOG_ERR, "MacGate: select() failed.");
			continue;
		}

		if(FD_ISSET(SockIP, &HearData))
		{
			from_len = sizeof(from);
                	if((bytes = recvfrom(SockIP, buf, sizeof( buf ), 0,
                		(struct sockaddr *) &from, &from_len )) < 0)
			{
                        	syslog(LOG_ERR, "MacGate: recvfrom SockIP");
			}
                	if(bytes > 0)
			{
				if(LTNodes != NULL)	/* Not on NULL list. */
                       			SendToLTalk(&LTNodes, buf, bytes);
 			}
		}

		if(FD_ISSET(SockLTIP, &HearData))
                {
                	from_len = sizeof(from);
                	if((bytes = recvfrom(SockLTIP, buf, sizeof( buf ), 0, 
				(struct sockaddr *) &from, &from_len )) < 0)
                	{
                		syslog(LOG_ERR, "MacGate: recvfrom SockLTIP");
                	}

			/* Make sure LT-IP packet. */
                	if(bytes > 0 && buf[SDDP_DDP_TYPE] == DDP_IP_TYPE)
                	{
                        	UpdateNodeMap(&LTNodes, buf);
                        	SendToInet(LTNodes, buf, bytes);
				// ExpireNodes(&LTNodes);
                	}
		}
#ifdef DEBUG_LIST
		PrintList(LTNodes);
#endif /* DEBUG_LIST */
        }
        return;
}

/*
 * Send the packet onwards to its final destination.. Mac Land. :)
 * We're called when a packet comes in off one of our pseudo ethernet
 * devices. We're sending DDP ShortHeader packets.
 */
void SendToLTalk (struct ltip_list **LTNodes, char buff[], int bytes)
{
	struct sockaddr target;
	struct ltip_list *RecvNode;	/* LocalTalk node we are sending to. */
	unsigned char full_packet[MAX_LLAP_SIZE];
	struct in_addr NodeIPNumber;
	short int lt_pkt_len, length;
	int Raw = 0;			/* Set to 1 if Raw IP packet. */

	/*
         * We give it 2 chances, Once check the Ethernet IP header
         * type and the other just the Raw IP packet. Seems we
         * get both types. (Raw IP I figure is PPP)
         */

	/* Destination IP of sender - Ethernet. */
        bcopy(&buff[DST_ETH_IP_NUM], &NodeIPNumber.s_addr, 4);
	RecvNode = ReturnNodeByIP(LTNodes, NodeIPNumber.s_addr);
	if(RecvNode == NULL)
	{ 	/* In case of Raw IP. */
        	bcopy(&buff[DST_ETH_IP_NUM-ETH_HDR_SIZE], 
			&NodeIPNumber.s_addr, 4);
		RecvNode = ReturnNodeByIP(LTNodes, NodeIPNumber.s_addr);
        	if(RecvNode == NULL)
			return;		/* Not one of our IP interfaces. */
		else
			Raw = 1;	/* We have pure/raw IP. */
	}

	/*
	 * Determine entire packet length. We need to watch for
	 * Ethernet IP packet or Raw IP packet.
	 */
	if(Raw)
		lt_pkt_len = (LLAP_HDRSIZE + SDDP_HDRSIZE) + bytes;
	else
		lt_pkt_len = (LLAP_HDRSIZE + SDDP_HDRSIZE) 
				+ (bytes - ETH_HDR_SIZE);

	/* LLAP Header. Easy */
        full_packet[0] 	= RecvNode->sat_addr.s_node; /* Destination node. */
        full_packet[1]	= MacGateNetNode.s_node;     /* LTalk Iface nodeid. */
        full_packet[LAP_TYPE] 	= 1;   		     /* Short DDP type. */

	/* Short DDP Header. Easy */
	length = ntohs(lt_pkt_len - LLAP_HDRSIZE);	  /* Length field. */
	bcopy(&length, &full_packet[SDDP_DDP_LENGTH], sizeof(length));
        full_packet[SDDP_DEST_SOCK]    	= MAC_IP_SOCKET;  /* Mac IP socket. */ 
        full_packet[SDDP_SRC_SOCK]     	= MAC_IP_SOCKET;  /* Our IP socket. */
        full_packet[SDDP_DDP_TYPE]     	= DDP_IP_TYPE; 	  /* IP DDP type. */

	/* 
	 * Copy IP Packet. Once again watching for either
	 * Ethernet IP or Raw IP.
	 */
	if(Raw)
		bcopy(&buff[0], &full_packet[LT_START_DATA], bytes);
	else
		bcopy(&buff[ETH_START_DATA], &full_packet[LT_START_DATA], 
		bytes-ETH_HDR_SIZE);

	bzero(&target, sizeof(struct sockaddr_at));
        target.sa_family = AF_INET;
        strcpy(target.sa_data, LT_DEV);  /* Name of device to send from. */

        /* Send packet upward and out. */
        if (sendto(SockLTIP, full_packet, lt_pkt_len, 0, 
		(struct sockaddr *) &target, sizeof( struct sockaddr_at )) < 0)
	{
                syslog(LOG_ERR, "MacGate: SendToLTalk - sendto() error.");
        }

        return;
}

/*
 * Send out the packet from the localtalk network. We ALWAYS
 * send to the localhost and then let it redirect the packet.
 * All we do is put an Ethernet IP type in the header and then
 * copy the IP packet over. Slick eh?
 */
void SendToInet (struct ltip_list *LTNodes, char buff[], int bytes)
{
	struct sockaddr target;
	unsigned char full_packet[MAX_LLAP_SIZE];
	int eth_pkt_len;

	/* Entire packet length in bytes. */
        eth_pkt_len = ETH_START_DATA + (bytes-(LLAP_HDRSIZE+SDDP_HDRSIZE));

	/* Ethernet Header, for Loopback no SRC/DST ethernet addrs. */
	bzero(&full_packet, MAX_LLAP_SIZE);
	full_packet[ETH_START_TYPE] = htons(ETH_P_IP);  /* Packet type. */

	/* IP Packet. */
        bcopy(&buff[LT_START_DATA], &full_packet[ETH_START_DATA], 
		bytes-(LLAP_HDRSIZE+SDDP_HDRSIZE));

	bzero(&target, sizeof(struct sockaddr_at));
	target.sa_family = AF_INET;
	strcpy(target.sa_data, LOOP_DEV);	/* Always sent via LoopBack. */

        /* Send packet upward and out. */
        if ( sendto(SockIP, full_packet, eth_pkt_len, 0, 
		(struct sockaddr *) &target, sizeof( struct sockaddr )) < 0 )
        {
                syslog(LOG_ERR, "MacGate: SendToInet - sendto() error.");
        }

        return;
}

/*
 * Our IP <-> LocalTalk routing table/list. Old entries will expire on
 * a timer. Dealing with a LocalTalk Short DDP packets here.
 * The pseudo ethernet interfaces are also controlled here.
 */
void UpdateNodeMap (struct ltip_list **LTNodes, char buff[])
{
        struct ltip_list *LTNewNode;
	struct in_addr NodeIPNumber;
	struct ltip_list *Node;

	/* Source IP of sender. (Macintosh) */
	bcopy(&buff[START_LT_IP_NUM], &NodeIPNumber.s_addr, 4);

	Node = ReturnNodeByIP(LTNodes, NodeIPNumber.s_addr);
	if(Node == NULL)
        {
                /* Malloc up a new ltip record. */
                LTNewNode = (struct ltip_list*)malloc(sizeof(struct ltip_list));
                LTNewNode->next = NULL;

                /* Set the Net:Node pair. */
                LTNewNode->sat_addr.s_node = buff[SRC_NODE];
                LTNewNode->sat_addr.s_net  = ATADDR_ANYNET; 
		
                /* Set the IP number for this node. */
                LTNewNode->ip = NodeIPNumber.s_addr;

                /* Set the time this route was created. */
                LTNewNode->time = time(NULL);

                /* Allocate a new ethernet interface for this node. */
                NewEtherNode(LTNodes, LTNewNode->ip, LTNewNode->name);
		
		/* Add/Update the node to our list. */
        	AddNode(LTNodes, LTNewNode);
	}
        return;
}

/*
 * Returns a stored routing entry by IP number.
 */
struct ltip_list* ReturnNodeByIP (struct ltip_list **LTNodes, __u32 IPNumber)
{
        struct ltip_list *current;

        if(*LTNodes == NULL)     /* Easy.. nothing there. */
                return NULL;    /* No match. */
        else
        {
                /* Check to see if it is at head of the list. */
                if((*LTNodes)->ip == IPNumber)
                        return *LTNodes;
                else
                {
                        current = (*LTNodes)->next;  /* Move to next one. */
                        while(current != NULL)
                        {
                                if(current->ip == IPNumber)
                                        return current;
                                else
                                        current = current->next;
                        }
                }
        }
        return NULL;
}

/*
 * Create a new PSEUDO ethernet/dummy device.
 * TODO: Check to see if system/user inserted a dev in our range. Otherwise
 * 	 we could take a used one, that we did not put there.
 */
void NewEtherNode (struct ltip_list **LTNodes,__u32 LTNewNodeIP,char name[])
{
        struct ltip_list *current;
	char buff[128];
	int x;

        if(*LTNodes == NULL)	/* None, we create first interface like this. */
		sprintf(name, "%s:%i", ETH_DEV, 0);
        else
        {       /* Find the next free pseudo ethernet device. */
                for(x=1; x<MAX_ETH_DEVS; x++)
                {
                        /* Throw together the dev name we want. */
			sprintf(name, "%s:%i", ETH_DEV, x);

                        current = (*LTNodes)->next;
                        while(current != NULL
				&& (strcmp(current->name, name) != 0))
			{
                        	current = current->next;
			}

                        if(current == NULL)     /* Yeah, we found a free one. */
			{
				/* Start up device. */
				bzero(&buff, 128);
        			sprintf(buff,"%s %s %s mtu %d", IFCONFIG, name,
                			in_ntoa(LTNewNodeIP), MTU);
        			system(buff);

				/* Add the route. */
        			bzero(&buff, 128);
        			sprintf(buff,"%s add -host %s dev %s", ROUTE, 
					in_ntoa(LTNewNodeIP), name);
        			system(buff);

				/* Add the firewall rule. */
        			bzero(&buff, 128);
        			sprintf(buff,"%s -I -a deny -D %s", IPFWADM, 
					in_ntoa(LTNewNodeIP));
				system(buff);

				NodesUp++; /* Add one to the Node counter. */

				return;
			}
                }
		return;	/* No devices available. */
        }
}

/*
 * Add a Localtalk IP node to the list.. we automaticly sort the
 * list as we add nodes. Sorting by IP number smallest to biggest.
 */
void AddNode (struct ltip_list **LTNodes, struct ltip_list *LTNewNode)
{
        struct ltip_list *current;
        struct ltip_list *shadow;

        if(*LTNodes == NULL)       /* If list is empty put at begining. */
                *LTNodes = LTNewNode;
        else
        {       /* Check to see if should be at head of list. */
                if(AddNodeBefore(**LTNodes, *LTNewNode))
                {
                        LTNewNode->next = *LTNodes;
                        *LTNodes = LTNewNode;
                }
                else    /* Insert somewhere not a begining. */
                {
                        shadow = *LTNodes;
                        current = (*LTNodes)->next;
                        while(current != NULL 
				&& !AddNodeBefore(*current, *LTNewNode))
                        {
                                shadow = current;
                                current = current->next;
                        }

                        /* Insert before current element */
                        shadow->next = LTNewNode;
                        LTNewNode->next = current;
                }
        }
        return;
}

/*
 * We look for a full match.
 * Node - Node; IP - IP;
 * TODO: Remove this... useless (merge)
 */
int AddNodeBefore (struct ltip_list LTNodeOne, struct ltip_list LTNodeTwo)
{
        /* Check to see if they are the same. */
        if((LTNodeOne.sat_addr.s_node == LTNodeTwo.sat_addr.s_node)
                && (LTNodeOne.ip == LTNodeTwo.ip))
                return 1;
        else
                return 0;       /* No Match. */
}

/*
 * Check all routes to see if any need to be deleted.
 */
void ExpireNodes (struct ltip_list **LTNodes)
{
        struct ltip_list *current;

        if(*LTNodes == NULL)    /* Easy.. nothing there. */
                return;         /* No routes. */
        else
        {
                current = *LTNodes;  /* Start at first route. */
                while(current != NULL)
                {
                        if((time(NULL) - current->time) > NODE_EXPIRE)
                                DeleteNode(LTNodes,*current); /* Delete route */
                        current = current->next;
                }
        }
        return;
}

/*
 * Delete a Localtalk IP node from the list and down
 * the Ethernet interface.
 */
void DeleteNode (struct ltip_list **LTNodes, struct ltip_list OldNode)
{
        struct ltip_list *current;
        struct ltip_list *shadow;

        if(*LTNodes == NULL)         /* Can't delete from empty list. */
                return;
        else
        {
                current = *LTNodes;
                if(AddNodeBefore(*current, OldNode)) /* Check beging. */
                        *LTNodes = current->next;
                else
                {       /* Find it in anywhere but in the begining. */
                        shadow = current;
                        current = current->next;
                        while((current != NULL) 
				&& (AddNodeBefore(*current, OldNode) != 1))
                        {
                                shadow = current;
                                current = current->next;
                        }
                        if(current == NULL)
                                return; /* Not found. */
                        else
                                shadow->next = current->next;
                }

                /* Delete item. */
                if(current != NULL)
                {
			/* Shutdown this node's Ethernet iface. */
                        EtherNodeDown(current->name);
                        free(current);
                }
        }
        return;
}

/*
 * Here we down the psuedo ethernet interface.
 */
void EtherNodeDown (char name[])
{
        char buf[128];
	struct ifreq ifr;
	struct sockaddr_in *s_in;
        int SockFD;

        SockFD = socket(AF_INET, SOCK_DGRAM, 0);
        if(SockFD < 0)
        {
                perror("MacGate: EtherNodeDown - socket(SockFD)");
                exit(1);
        }

        /* Get the IP so we can kill the IP firewall rule. */
        strcpy(ifr.ifr_name, name);
        if (ioctl(SockFD, SIOCGIFADDR, &ifr) < 0) perror("SIOCGIFADDR");
        if (ioctl(SockFD, SIOCGIFADDR, &ifr) < 0) perror("SIOCGIFADDR");
        s_in = (struct sockaddr_in *)&ifr.ifr_addr;

	bzero(&buf, 128);
        sprintf(buf,"%s -I -d deny -D %s", IPFWADM, 
		in_ntoa(s_in->sin_addr.s_addr));
	system(buf);

	bzero(&buf, 128);
        sprintf(buf,"%s %s down", IFCONFIG, name);
        system(buf);

        return;
}

/*
 * Close down all psuedo ethernet interfaces.
 */
void AllEtherDown (void)
{
	char name[6];
	int x;
	
	for(x=PSEUDO_START; x<=NodesUp; x++)
	{
		/* Throw together the dev name we want. */
		sprintf(name, "%s:%i", ETH_DEV, x);

		EtherNodeDown(name);
	}

	return;
}
