/*
 * MacGate.h - Header for MacGate a LocalTalk to IP Gateway for Linux.
 *	Written By: Jay Schulist <Jay.Schulist@spacs.k12.wi.us>
 *	Copyright (C) 1997 Jay Schulist
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <netatalk/endian.h>
#include <netatalk/at.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <signal.h>
#include <unistd.h>
#include <linux/socket.h>
#include <atalk/atp.h>
#include <atalk/nbp.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/fcntl.h>
#include <linux/if.h>
#include <linux/if_ether.h>
#include <arpa/inet.h>  /* for inet_aton */
#include <netdb.h>

#ifndef MAX
#define MAX(a, b)       ((a) > (b) ? (a) : (b))
#endif

#define MAX_IFS         100

#define FLAGS_GOOD (IFF_UP          | IFF_BROADCAST)
#define FLAGS_MASK (IFF_UP          | IFF_BROADCAST | \
                    IFF_POINTOPOINT | IFF_LOOPBACK  | IFF_NOARP)

/* Paths to program locations. */
#define IFCONFIG	"/sbin/ifconfig"	/* Path to ifconfig. */
#define ROUTE		"/sbin/route"		/* Path to route. */
#define IPFWADM		"/sbin/ipfwadm"		/* Path to ipfwadm. */

/* Device/system specific defines. */
#define LT_DEV		"lt0"           /* LocalTalk device to use. */
#define ETH_DEV         "eth0"		/* Ethernet device to use. */
#define LOOP_DEV	"lo"		/* LoopBack device to use. */
#define PSEUDO_START      0             /* Pseudo's start here up. */
#define MAX_ETH_DEVS    255             /* Maximum Pseudo ethernet devices. */
#define NODE_EXPIRE      60             /* Seconds route has to live after packet passes. */

/*
 * There is no need to change anything below here.
 */

/* Ethernet specific defines. */
#define ETH_HDR_SIZE	14		/* Size of ethernet header. */
#define ETH_START_DATA	14		/* Where we can start putting data. */
#define ETH_START_TYPE  12		/* Ethernet packet type field. */
#define MTU             585             /* MTU. */

#define SRC_ETH_IP_NUM		26
#define DST_ETH_IP_NUM		30	/* Where we start pulling IP num from. */
#define START_LT_IP_NUM		20	/* Where we start pulling IP num from. */

#define MAC_IP_SOCKET	72		/* Macintosh IP socket. */
#define DDP_IP_TYPE	22		/* DDP IP Code. */

/* LLAP Header information. */
#define LLAP_HDRSIZE	  3		/* Length of LLAP header. */
#define MAX_LLAP_SIZE	603		/* Including LAP header and data. */
#define MAX_LAP_DATA	600		/* Max LAP Data size - all LAPs. */

/* LLAP Header offsets. */
/* 
 * Seems SOCK_PACKET adds 3 bytes and messes a few other things...
 * we will work around it for now.
 */
#define DEST_NODE	  3		/* Destination node ID. */
#define SRC_NODE	  4		/* Source node ID. */
#define LAP_TYPE	  2		/* LLAP type.  Short or Extended DDP Header */

/* Short DDP Header information. */
#define SDDP_HDRSIZE	  5		/* Length of Short DDP header bytes. */
#define SDDP_DDP_LENGTH	  3		/* Start of 2 byte length field. */
#define SDDP_DEST_SOCK	  5		/* Destination Socket Number. */
#define SDDP_SRC_SOCK	  6		/* Source Socket Number. */
#define SDDP_DDP_TYPE	  7		/* DDP Type. */

/* Packet Data information. */
#define MAX_DATA_SIZE	585		/* Maximum bytes data. */
#define LT_START_DATA	  8		/* Data field start. */

struct ltip_list {
        struct ltip_list *next;		/* Next record. */
        struct at_addr sat_addr;        /* LocalTalk Net:Node pair. */
	unsigned char name[6];		/* Device name. */
        __u32 ip;                       /* IP address of Macintosh. */
        long time;                      /* Time packet was recieved. */
};
